@@ -47,6 +47,13 @@ const int rpmFLAGS = RPMSENSE_EQUAL;
 #undef HTKEYTYPE
 #undef HTDATATYPE
 
+#define HASHTYPE conflictsCache
+#define HTKEYTYPE const char *
+#include "rpmhash.H"
+#include "rpmhash.C"
+#undef HASHTYPE
+#undef HTKEYTYPE
+
 /**
  * Check for supported payload format in header.
  * @param h		header to check
@@ -568,23 +575,31 @@ retry:
 
     /* Pretrans dependencies can't be satisfied by added packages. */
     if (!(dsflags & RPMSENSE_PRETRANS)) {
-	rpmte match = rpmalSatisfiesDepend(tsmem->addedPackages, dep);
+	rpmte *matches = rpmalAllSatisfiesDepend(tsmem->addedPackages, dep);
+	int match = 0;
 
 	/*
 	 * Handle definitive matches within the added package set.
 	 * Self-obsoletes and -conflicts fall through here as we need to 
 	 * check for possible other matches in the rpmdb.
 	 */
-	if (match) {
+	for (rpmte *m = matches; m && *m; m++) {
 	    rpmTagVal dtag = rpmdsTagN(dep);
 	    /* Requires match, look no further */
-	    if (dtag == RPMTAG_REQUIRENAME)
-		goto exit;
+	    if (dtag == RPMTAG_REQUIRENAME) {
+		match = 1;
+		break;
+	    }
 
 	    /* Conflicts/obsoletes match on another package, look no further */
-	    if (rpmteDS(match, dtag) != dep)
-		goto exit;
+	    if (rpmteDS(*m, dtag) != dep) {
+		match = 1;
+		break;
+	    }
 	}
+	free(matches);
+	if (match)
+	    goto exit;
     }
 
     /* See if the rpmdb provides it */
@@ -668,6 +683,7 @@ int rpmtsCheck(rpmts ts)
     int closeatexit = 0;
     int rc = 0;
     depCache dcache = NULL;
+    conflictsCache confcache = NULL;
     
     (void) rpmswEnter(rpmtsOp(ts, RPMTS_OP_CHECK), 0);
 
@@ -682,6 +698,27 @@ int rpmtsCheck(rpmts ts)
     dcache = depCacheCreate(5001, rstrhash, strcmp,
 				     (depCacheFreeKey)rfree, NULL);
 
+    confcache = conflictsCacheCreate(257, rstrhash, strcmp,
+				     (depCacheFreeKey)rfree);
+    if (confcache) {
+	rpmdbIndexIterator ii = rpmdbIndexIteratorInit(rpmtsGetRdb(ts), RPMTAG_CONFLICTNAME);
+	if (ii) {
+	    char *key;
+	    size_t keylen;
+	    while ((rpmdbIndexIteratorNext(ii, (const void**)&key, &keylen)) == 0) {
+		char *k;
+		if (!key || keylen == 0 || key[0] != '/')
+		    continue;
+		k = rmalloc(keylen + 1);
+		memcpy(k, key, keylen);
+		k[keylen] = 0;
+		conflictsCacheAddEntry(confcache, k);
+	    }
+	    rpmdbIndexIteratorFree(ii);
+	}
+    }
+
+    
     /*
      * Look at all of the added packages and make sure their dependencies
      * are satisfied.
@@ -711,6 +748,17 @@ int rpmtsCheck(rpmts ts)
 
 	/* Check package name (not provides!) against installed obsoletes */
 	checkInstDeps(ts, dcache, p, RPMTAG_OBSOLETENAME, rpmteN(p));
+
+	/* Check filenames against installed conflicts */
+        if (conflictsCacheNumKeys(confcache)) {
+	    rpmfi fi = rpmfiInit(rpmteFI(p), 0);
+	    while (rpmfiNext(fi) >= 0) {
+		const char *fn = rpmfiFN(fi);
+		if (!conflictsCacheHasEntry(confcache, fn))
+		    continue;
+		checkInstDeps(ts, dcache, p, RPMTAG_CONFLICTNAME, fn);
+	    }
+	}
     }
     rpmtsiFree(pi);
 
@@ -739,6 +787,7 @@ int rpmtsCheck(rpmts ts)
 
 exit:
     depCacheFree(dcache);
+    conflictsCacheFree(confcache);
 
     (void) rpmswExit(rpmtsOp(ts, RPMTS_OP_CHECK), 0);
 
