@@ -121,6 +121,7 @@ static struct tableType_s tables[RPM_MACHTABLE_COUNT] = {
 /* XXX get rid of this stuff... */
 /* Stuff for maintaining "variables" like SOURCEDIR, BUILDDIR, etc */
 #define RPMVAR_OPTFLAGS                 3
+#define RPMVAR_ARCHCOLOR                42
 #define RPMVAR_INCLUDE                  43
 #define RPMVAR_MACROFILES               49
 
@@ -130,6 +131,7 @@ static struct tableType_s tables[RPM_MACHTABLE_COUNT] = {
 /* The order of the flags is archSpecific, macroize, localize */
 
 static const struct rpmOption optionTable[] = {
+    { "archcolor",		RPMVAR_ARCHCOLOR,               1, 0, 0 },
     { "include",		RPMVAR_INCLUDE,			0, 0, 2 },
     { "macrofiles",		RPMVAR_MACROFILES,		0, 0, 1 },
     { "optflags",		RPMVAR_OPTFLAGS,		1, 1, 0 },
@@ -435,6 +437,7 @@ static void setDefaults(void)
 #ifndef MACROFILES
     if (!macrofiles) {
 	macrofiles = rstrscat(NULL, confdir, "/macros", ":",
+				confdir, "/macros.d/macros.*", ":",
 				confdir, "/platform/%{_target}/macros", ":",
 				confdir, "/fileattrs/*.attr", ":",
   				confdir, "/" RPMCANONVENDOR "/macros", ":",
@@ -1354,6 +1357,28 @@ void rpmGetArchInfo(const char ** name, int * num)
     getMachineInfo(ARCH, name, num);
 }
 
+int rpmGetArchColor(const char *arch)
+{
+    const char *color;
+    char *e;
+    int color_i;
+
+    arch = lookupInDefaultTable(arch,
+				tables[currTables[ARCH]].defaults,
+				tables[currTables[ARCH]].defaultsLength);
+    color = rpmGetVarArch(RPMVAR_ARCHCOLOR, arch);
+    if (color == NULL) {
+	return -1;
+    }
+
+    color_i = strtol(color, &e, 10);
+    if (!(e && *e == '\0')) {
+	return -1;
+    }
+
+    return color_i;
+}
+
 void rpmGetOsInfo(const char ** name, int * num)
 {
     getMachineInfo(OS, name, num);
@@ -1582,7 +1607,8 @@ int rpmReadConfigFiles(const char * file, const char * target)
 {
     /* Force preloading of dlopen()'ed libraries in case we go chrooting */
     (void) gethostbyname("localhost");
-    (void) rpmInitCrypto();
+    if (rpmInitCrypto())
+	return -1;
 
     /* Preset target macros */
    	/* FIX: target can be NULL */
@@ -1678,6 +1704,9 @@ int rpmShowRC(FILE * fp)
     ds = rpmdsFree(ds);
     fprintf(fp, "\n");
 
+    fprintf(fp, "Macro path: %s\n", macrofiles);
+    fprintf(fp, "\n");
+
     rpmDumpMacroTable(NULL, fp);
 
     return 0;
