@@ -15,6 +15,10 @@
 
 #include "lib/rpmplugins.h"
 #include "lib/rpmte_internal.h"
+/* strpool-related interfaces */
+#include "lib/rpmfi_internal.h"
+#include "lib/rpmds_internal.h"
+#include "lib/rpmts_internal.h"
 
 #include "debug.h"
 
@@ -94,12 +98,12 @@ static rpmfi getFI(rpmte p, Header h)
 				      (RPMFI_NOHEADER | RPMFI_FLAGS_ERASE);
 
     /* relocate stuff in header if necessary */
-    if (rpmteType(p) == TR_ADDED && rpmfsFC(p->fs) > 0) {
+    if (rpmteType(p) == TR_ADDED && rpmfsFC(p->fs) > 0 && p->nrelocs) {
 	if (!headerIsSource(h) && !headerIsEntry(h, RPMTAG_ORIGBASENAMES)) {
 	    rpmRelocateFileList(p->relocs, p->nrelocs, p->fs, h);
 	}
     }
-    return rpmfiNew(NULL, h, RPMTAG_BASENAMES, fiflags);
+    return rpmfiNewPool(rpmtsPool(p->ts), h, RPMTAG_BASENAMES, fiflags);
 }
 
 /* stupid bubble sort, but it's probably faster here */
@@ -203,7 +207,8 @@ static void buildRelocs(rpmte p, Header h, rpmRelocation *relocs)
  */
 static int addTE(rpmte p, Header h, fnpyKey key, rpmRelocation * relocs)
 {
-    struct rpmtd_s colls;
+    rpmstrPool tspool = rpmtsPool(p->ts);
+    struct rpmtd_s colls, bnames;
     int rc = 1; /* assume failure */
 
     p->name = headerGetAsString(h, RPMTAG_NAME);
@@ -241,14 +246,18 @@ static int addTE(rpmte p, Header h, fnpyKey key, rpmRelocation * relocs)
     p->pkgFileSize = 0;
     p->headerSize = headerSizeof(h, HEADER_MAGIC_NO);
 
-    p->thisds = rpmdsThis(h, RPMTAG_PROVIDENAME, RPMSENSE_EQUAL);
-    p->provides = rpmdsNew(h, RPMTAG_PROVIDENAME, 0);
-    p->requires = rpmdsNew(h, RPMTAG_REQUIRENAME, 0);
-    p->conflicts = rpmdsNew(h, RPMTAG_CONFLICTNAME, 0);
-    p->obsoletes = rpmdsNew(h, RPMTAG_OBSOLETENAME, 0);
-    p->order = rpmdsNew(h, RPMTAG_ORDERNAME, 0);
+    p->thisds = rpmdsThisPool(tspool, h, RPMTAG_PROVIDENAME, RPMSENSE_EQUAL);
+    p->provides = rpmdsNewPool(tspool, h, RPMTAG_PROVIDENAME, 0);
+    p->requires = rpmdsNewPool(tspool, h, RPMTAG_REQUIRENAME, 0);
+    p->conflicts = rpmdsNewPool(tspool, h, RPMTAG_CONFLICTNAME, 0);
+    p->obsoletes = rpmdsNewPool(tspool, h, RPMTAG_OBSOLETENAME, 0);
+    p->order = rpmdsNewPool(tspool, h, RPMTAG_ORDERNAME, 0);
+
+    /* Relocation needs to know file count before rpmfiNew() */
+    headerGet(h, RPMTAG_BASENAMES, &bnames, HEADERGET_MINMEM);
+    p->fs = rpmfsNew(rpmtdCount(&bnames), (p->type == TR_ADDED));
+    rpmtdFreeData(&bnames);
 
-    p->fs = rpmfsNew(h, p->type);
     p->fi = getFI(p, h);
 
     /* Packages with no files return an empty file info set, NULL is an error */
@@ -768,7 +777,7 @@ int rpmteFailed(rpmte te)
     return (te != NULL) ? te->failed : -1;
 }
 
-static int rpmteHaveTransScript(rpmte te, rpmTagVal tag)
+int rpmteHaveTransScript(rpmte te, rpmTagVal tag)
 {
     int rc = 0;
     if (tag == RPMTAG_PRETRANS) {
