@@ -1216,7 +1216,9 @@ ssize_t Fread(void *buf, size_t size, size_t nmemb, FD_t fd)
 	fdio_read_function_t _read = FDIOVEC(fd, read);
 
 	fdstat_enter(fd, FDSTAT_READ);
-	rc = (_read ? (*_read) (fd, buf, size * nmemb) : -2);
+	do {
+	    rc = (_read ? (*_read) (fd, buf, size * nmemb) : -2);
+	} while (rc == -1 && errno == EINTR);
 	fdstat_exit(fd, FDSTAT_READ, rc);
 
 	if (fd->digests && rc > 0)
@@ -1237,7 +1239,9 @@ ssize_t Fwrite(const void *buf, size_t size, size_t nmemb, FD_t fd)
 	fdio_write_function_t _write = FDIOVEC(fd, write);
 	
 	fdstat_enter(fd, FDSTAT_WRITE);
-	rc = (_write ? _write(fd, buf, size * nmemb) : -2);
+	do {
+	    rc = (_write ? _write(fd, buf, size * nmemb) : -2);
+	} while (rc == -1 && errno == EINTR);
 	fdstat_exit(fd, FDSTAT_WRITE, rc);
 
 	if (fd->digests && rc > 0)
@@ -1370,7 +1374,7 @@ static void cvtfmode (const char *m,
 
 FD_t Fdopen(FD_t ofd, const char *fmode)
 {
-    char stdio[20], other[20], zstdio[20];
+    char stdio[20], other[20], zstdio[40];
     const char *end = NULL;
     FDIO_t iof = NULL;
     FD_t fd = ofd;
@@ -1385,8 +1389,8 @@ fprintf(stderr, "*** Fdopen(%p,%s) %s\n", fd, fmode, fdbg(fd));
     if (stdio[0] == '\0')
 	return NULL;
     zstdio[0] = '\0';
-    strncat(zstdio, stdio, sizeof(zstdio) - strlen(zstdio));
-    strncat(zstdio, other, sizeof(zstdio) - strlen(zstdio));
+    strncat(zstdio, stdio, sizeof(zstdio) - strlen(zstdio) - 1);
+    strncat(zstdio, other, sizeof(zstdio) - strlen(zstdio) - 1);
 
     if (end == NULL && other[0] == '\0')
 	return fd;
