@@ -166,13 +166,13 @@ int main(int argc, char *argv[])
     if (ia->prefix && ia->prefix[0] != '/') 
 	argerror(_("arguments to --prefix must begin with a /"));
 
-    if (bigMode != MODE_INSTALL && (ia->installInterfaceFlags & INSTALL_HASH))
+    if (!(bigMode & MODES_IE) && (ia->installInterfaceFlags & INSTALL_HASH))
 	argerror(_("--hash (-h) may only be specified during package "
-			"installation"));
+			"installation and erasure"));
 
-    if (bigMode != MODE_INSTALL && (ia->installInterfaceFlags & INSTALL_PERCENT))
+    if (!(bigMode & MODES_IE) && (ia->installInterfaceFlags & INSTALL_PERCENT))
 	argerror(_("--percent may only be specified during package "
-			"installation"));
+			"installation and erasure"));
 
     if (bigMode != MODE_INSTALL && (ia->probFilter & RPMPROB_FILTER_REPLACEPKG))
 	argerror(_("--replacepkgs may only be specified during package "
@@ -332,18 +332,20 @@ int main(int argc, char *argv[])
 	break;
     }
 
-    ts = rpmtsFree(ts);
-    finishPipe();
+    rpmtsFree(ts);
+    if (finishPipe())
+	ec = EXIT_FAILURE;
 
 #ifdef	IAM_RPMQV
-    qva->qva_queryFormat = _free(qva->qva_queryFormat);
+    free(qva->qva_queryFormat);
 #endif
 
 #ifdef	IAM_RPMEIU
-    if (ia->relocations != NULL)
-    for (i = 0; i < ia->numRelocations; i++)
-	ia->relocations[i].oldPath = _free(ia->relocations[i].oldPath);
-    ia->relocations = _free(ia->relocations);
+    if (ia->relocations != NULL) {
+	for (i = 0; i < ia->numRelocations; i++)
+	    free(ia->relocations[i].oldPath);
+	free(ia->relocations);
+    }
 #endif
 
     rpmcliFini(optCon);
