@@ -17,7 +17,6 @@
 #include <rpm/rpmlog.h>
 #include <rpm/rpmstring.h>
 
-#include "rpmio/digest.h"
 #include "lib/rpmlead.h"
 #include "lib/signature.h"
 
@@ -115,22 +114,26 @@ exit:
  */
 static int putSignature(Header sigh, int ishdr, uint8_t *pkt, size_t pktlen)
 {
-    pgpDig dig = pgpNewDig();
-    pgpDigParams sigp = &dig->signature;
+    pgpDigParams sigp = NULL;
     rpmTagVal sigtag;
     struct rpmtd_s sigtd;
     int rc = 1; /* assume failure */
+    unsigned int hash_algo;
+    unsigned int pubkey_algo;
 
-    if (pgpPrtPkts(pkt, pktlen, dig, 0) != 0)
-	goto exit;
-
-    if (rpmDigestLength(sigp->hash_algo) == 0) {
-	rpmlog(RPMLOG_ERR, _("Unsupported PGP hash algorithm %d\n"),
-	       sigp->hash_algo);
+    if (pgpPrtParams(pkt, pktlen, PGPTAG_SIGNATURE, &sigp)) {
+	rpmlog(RPMLOG_ERR, _("Unsupported PGP signature\n"));
 	goto exit;
     }
 
-    switch (sigp->pubkey_algo) {
+    hash_algo = pgpDigParamsAlgo(sigp, PGPVAL_HASHALGO);
+    if (rpmDigestLength(hash_algo) == 0) {
+	rpmlog(RPMLOG_ERR, _("Unsupported PGP hash algorithm %u\n"), hash_algo);
+	goto exit;
+    }
+
+    pubkey_algo = pgpDigParamsAlgo(sigp, PGPVAL_PUBKEYALGO);
+    switch (pubkey_algo) {
     case PGPPUBKEYALGO_DSA:
 	sigtag = ishdr ? RPMSIGTAG_DSA : RPMSIGTAG_GPG;
 	break;
@@ -138,8 +141,8 @@ static int putSignature(Header sigh, int ishdr, uint8_t *pkt, size_t pktlen)
 	sigtag = ishdr ? RPMSIGTAG_RSA : RPMSIGTAG_PGP;
 	break;
     default:
-	rpmlog(RPMLOG_ERR, _("Unsupported PGP pubkey algorithm %d\n"),
-		sigp->pubkey_algo);
+	rpmlog(RPMLOG_ERR, _("Unsupported PGP pubkey algorithm %u\n"),
+		pubkey_algo);
 	goto exit;
 	break;
     }
@@ -155,7 +158,7 @@ static int putSignature(Header sigh, int ishdr, uint8_t *pkt, size_t pktlen)
     rc = (headerPut(sigh, &sigtd, HEADERPUT_DEFAULT) == 0);
 
 exit:
-    dig = pgpFreeDig(dig);
+    pgpDigParamsFree(sigp);
     return rc;
 }
 
@@ -310,7 +313,7 @@ exit:
 	(void) unlink(fn);
 	free(fn);
     }
-    h = headerFree(h);
+    headerFree(h);
     if (fd != NULL) (void) Fclose(fd);
     return ret;
 }
@@ -334,20 +337,16 @@ static int rpmGenSignature(Header sigh, const char * file,
  * @param sigtag	signature tag
  * @return		parsed pgp dig or NULL
  */
-static pgpDig getSig(Header sigh, rpmTagVal sigtag)
+static pgpDigParams getSig(Header sigh, rpmTagVal sigtag)
 {
     struct rpmtd_s pkt;
-    pgpDig dig = NULL;
+    pgpDigParams sig = NULL;
 
     if (headerGet(sigh, sigtag, &pkt, HEADERGET_DEFAULT) && pkt.data != NULL) {
-	dig = pgpNewDig();
-
-	if (pgpPrtPkts(pkt.data, pkt.count, dig, 0) != 0) {
-	    dig = pgpFreeDig(dig);
-	}
+	pgpPrtParams(pkt.data, pkt.count, PGPTAG_SIGNATURE, &sig);
 	rpmtdFreeData(&pkt);
     }
-    return dig;
+    return sig;
 }
 
 static void deleteSigs(Header sigh)
@@ -361,35 +360,14 @@ static void deleteSigs(Header sigh)
 
 static int sameSignature(rpmTagVal sigtag, Header h1, Header h2)
 {
-    pgpDig dig1 = getSig(h1, sigtag);
-    pgpDig dig2 = getSig(h2, sigtag);
-    int rc = 0; /* assume different, eg if either signature doesn't exist */
+    pgpDigParams sig1 = getSig(h1, sigtag);
+    pgpDigParams sig2 = getSig(h2, sigtag);;
 
-    /* XXX This part really belongs to rpmpgp.[ch] */
-    if (dig1 && dig2) {
-	pgpDigParams sig1 = &dig1->signature;
-	pgpDigParams sig2 = &dig2->signature;
+    int rc = pgpDigParamsCmp(sig1, sig2);
 
-	/* XXX Should we compare something else too? */
-	if (sig1->hash_algo != sig2->hash_algo)
-	    goto exit;
-	if (sig1->pubkey_algo != sig2->pubkey_algo)
-	    goto exit;
-	if (sig1->version != sig2->version)
-	    goto exit;
-	if (sig1->sigtype != sig2->sigtype)
-	    goto exit;
-	if (memcmp(sig1->signid, sig2->signid, sizeof(sig1->signid)) != 0)
-	    goto exit;
-
-	/* Parameters match, assume same signature */
-	rc = 1;
-    }
-
-exit:
-    pgpFreeDig(dig1);
-    pgpFreeDig(dig2);
-    return rc;
+    pgpDigParamsFree(sig1);
+    pgpDigParamsFree(sig2);
+    return (rc == 0);
 }
 
 static int replaceSignature(Header sigh, const char *sigtarget,
@@ -430,175 +408,156 @@ static int rpmSign(const char *rpm, int deleting, const char *passPhrase)
 {
     FD_t fd = NULL;
     FD_t ofd = NULL;
-    rpmlead lead;
+    rpmlead lead = NULL;
     char *sigtarget = NULL, *trpm = NULL;
     Header sigh = NULL;
-    char * msg;
+    char * msg = NULL;
     int res = -1; /* assume failure */
-    int xx;
-    
-    {
-    	rpmRC rc;
-	struct rpmtd_s utd;
+    rpmRC rc;
+    struct rpmtd_s utd;
 
-	fprintf(stdout, "%s:\n", rpm);
+    fprintf(stdout, "%s:\n", rpm);
 
-	if (manageFile(&fd, rpm, O_RDONLY))
-	    goto exit;
+    if (manageFile(&fd, rpm, O_RDONLY))
+	goto exit;
 
-	lead = rpmLeadNew();
+    if ((rc = rpmLeadRead(fd, &lead, NULL, &msg)) != RPMRC_OK) {
+	rpmlog(RPMLOG_ERR, "%s: %s\n", rpm, msg);
+	free(msg);
+	goto exit;
+    }
 
-	if ((rc = rpmLeadRead(fd, lead)) == RPMRC_OK) {
-	    const char *lmsg = NULL;
-	    rc = rpmLeadCheck(lead, &lmsg);
-	    if (rc != RPMRC_OK) 
-		rpmlog(RPMLOG_ERR, "%s: %s\n", rpm, lmsg);
-	}
-
-	if (rc != RPMRC_OK) {
-	    lead = rpmLeadFree(lead);
-	    goto exit;
-	}
-
-	msg = NULL;
-	rc = rpmReadSignature(fd, &sigh, RPMSIGTYPE_HEADERSIG, &msg);
-	switch (rc) {
-	default:
-	    rpmlog(RPMLOG_ERR, _("%s: rpmReadSignature failed: %s"), rpm,
-			(msg && *msg ? msg : "\n"));
-	    msg = _free(msg);
-	    goto exit;
-	    break;
-	case RPMRC_OK:
-	    if (sigh == NULL) {
-		rpmlog(RPMLOG_ERR, _("%s: No signature available\n"), rpm);
-		goto exit;
-	    }
-	    break;
-	}
+    rc = rpmReadSignature(fd, &sigh, RPMSIGTYPE_HEADERSIG, &msg);
+    switch (rc) {
+    default:
+	rpmlog(RPMLOG_ERR, _("%s: rpmReadSignature failed: %s"), rpm,
+		    (msg && *msg ? msg : "\n"));
 	msg = _free(msg);
-
-	/* ASSERT: ofd == NULL && sigtarget == NULL */
-	ofd = rpmMkTempFile(NULL, &sigtarget);
-	if (ofd == NULL || Ferror(ofd)) {
-	    rpmlog(RPMLOG_ERR, _("rpmMkTemp failed\n"));
+	goto exit;
+	break;
+    case RPMRC_OK:
+	if (sigh == NULL) {
+	    rpmlog(RPMLOG_ERR, _("%s: No signature available\n"), rpm);
 	    goto exit;
 	}
-	/* Write the header and archive to a temp file */
-	if (copyFile(&fd, rpm, &ofd, sigtarget))
-	    goto exit;
-	/* Both fd and ofd are now closed. sigtarget contains tempfile name. */
-	/* ASSERT: fd == NULL && ofd == NULL */
+	break;
+    }
+    msg = _free(msg);
 
-	/* Dump the immutable region (if present). */
-	if (headerGet(sigh, RPMTAG_HEADERSIGNATURES, &utd, HEADERGET_DEFAULT)) {
-	    struct rpmtd_s copytd;
-	    Header nh = headerNew();
-	    Header oh = headerCopyLoad(utd.data);
-	    HeaderIterator hi = headerInitIterator(oh);
-	    while (headerNext(hi, &copytd)) {
-		if (copytd.data)
-		    xx = headerPut(nh, &copytd, HEADERPUT_DEFAULT);
-		rpmtdFreeData(&copytd);
-	    }
-	    hi = headerFreeIterator(hi);
-	    oh = headerFree(oh);
+    ofd = rpmMkTempFile(NULL, &sigtarget);
+    if (ofd == NULL || Ferror(ofd)) {
+	rpmlog(RPMLOG_ERR, _("rpmMkTemp failed\n"));
+	goto exit;
+    }
+    /* Write the header and archive to a temp file */
+    if (copyFile(&fd, rpm, &ofd, sigtarget))
+	goto exit;
+    /* Both fd and ofd are now closed. sigtarget contains tempfile name. */
 
-	    sigh = headerFree(sigh);
-	    sigh = headerLink(nh);
-	    nh = headerFree(nh);
+    /* Dump the immutable region (if present). */
+    if (headerGet(sigh, RPMTAG_HEADERSIGNATURES, &utd, HEADERGET_DEFAULT)) {
+	struct rpmtd_s copytd;
+	Header nh = headerNew();
+	Header oh = headerCopyLoad(utd.data);
+	HeaderIterator hi = headerInitIterator(oh);
+	while (headerNext(hi, &copytd)) {
+	    if (copytd.data)
+		headerPut(nh, &copytd, HEADERPUT_DEFAULT);
+	    rpmtdFreeData(&copytd);
 	}
+	headerFreeIterator(hi);
+	headerFree(oh);
 
-	/* Eliminate broken digest values. */
-	xx = headerDel(sigh, RPMSIGTAG_BADSHA1_1);
-	xx = headerDel(sigh, RPMSIGTAG_BADSHA1_2);
+	headerFree(sigh);
+	sigh = headerLink(nh);
+	headerFree(nh);
+    }
 
-	/* Toss and recalculate header+payload size and digests. */
-	{
-	    rpmTagVal const sigs[] = { 	RPMSIGTAG_SIZE, 
-					RPMSIGTAG_MD5,
-				  	RPMSIGTAG_SHA1,
-				     };
-	    int nsigs = sizeof(sigs) / sizeof(rpmTagVal);
-	    for (int i = 0; i < nsigs; i++) {
-		(void) headerDel(sigh, sigs[i]);
-		if (rpmGenDigest(sigh, sigtarget, sigs[i]))
-		    goto exit;
-	    }
-	}
+    /* Eliminate broken digest values. */
+    headerDel(sigh, RPMSIGTAG_BADSHA1_1);
+    headerDel(sigh, RPMSIGTAG_BADSHA1_2);
 
-	if (deleting) {	/* Nuke all the signature tags. */
-	    deleteSigs(sigh);
-	} else {
-	    res = replaceSignature(sigh, sigtarget, passPhrase);
-	    if (res != 0) {
-		if (res == 1) {
-		    rpmlog(RPMLOG_WARNING,
-		       _("%s already contains identical signature, skipping\n"),
-		       rpm);
-		    /* Identical signature is not an error */
-		    res = 0;
-		}
+    /* Toss and recalculate header+payload size and digests. */
+    {
+	rpmTagVal const sigs[] = { 	RPMSIGTAG_SIZE, 
+				    RPMSIGTAG_MD5,
+				    RPMSIGTAG_SHA1,
+				 };
+	int nsigs = sizeof(sigs) / sizeof(rpmTagVal);
+	for (int i = 0; i < nsigs; i++) {
+	    (void) headerDel(sigh, sigs[i]);
+	    if (rpmGenDigest(sigh, sigtarget, sigs[i]))
 		goto exit;
-	    }
-	}
-
-	/* Reallocate the signature into one contiguous region. */
-	sigh = headerReload(sigh, RPMTAG_HEADERSIGNATURES);
-	if (sigh == NULL)	/* XXX can't happen */
-	    goto exit;
-
-	rasprintf(&trpm, "%s.XXXXXX", rpm);
-	ofd = rpmMkTemp(trpm);
-	if (ofd == NULL || Ferror(ofd)) {
-	    rpmlog(RPMLOG_ERR, _("rpmMkTemp failed\n"));
-	    goto exit;
-	}
-
-	/* Write the lead/signature of the output rpm */
-	rc = rpmLeadWrite(ofd, lead);
-	lead = rpmLeadFree(lead);
-	if (rc != RPMRC_OK) {
-	    rpmlog(RPMLOG_ERR, _("%s: writeLead failed: %s\n"), trpm,
-		Fstrerror(ofd));
-	    goto exit;
-	}
-
-	if (rpmWriteSignature(ofd, sigh)) {
-	    rpmlog(RPMLOG_ERR, _("%s: rpmWriteSignature failed: %s\n"), trpm,
-		Fstrerror(ofd));
-	    goto exit;
-	}
-
-	/* Append the header and archive from the temp file */
-	/* ASSERT: fd == NULL && ofd != NULL */
-	if (copyFile(&fd, sigtarget, &ofd, trpm))
-	    goto exit;
-	/* Both fd and ofd are now closed. */
-	/* ASSERT: fd == NULL && ofd == NULL */
-
-	/* Move final target into place, restore file permissions. */
-	{
-	    struct stat st;
-	    xx = stat(rpm, &st);
-	    xx = unlink(rpm);
-	    xx = rename(trpm, rpm);
-	    xx = chmod(rpm, st.st_mode);
 	}
     }
 
-    res = 0;
+    if (deleting) {	/* Nuke all the signature tags. */
+	deleteSigs(sigh);
+    } else {
+	res = replaceSignature(sigh, sigtarget, passPhrase);
+	if (res != 0) {
+	    if (res == 1) {
+		rpmlog(RPMLOG_WARNING,
+		   _("%s already contains identical signature, skipping\n"),
+		   rpm);
+		/* Identical signature is not an error */
+		res = 0;
+	    }
+	    goto exit;
+	}
+    }
+
+    /* Reallocate the signature into one contiguous region. */
+    sigh = headerReload(sigh, RPMTAG_HEADERSIGNATURES);
+    if (sigh == NULL)	/* XXX can't happen */
+	goto exit;
+
+    rasprintf(&trpm, "%s.XXXXXX", rpm);
+    ofd = rpmMkTemp(trpm);
+    if (ofd == NULL || Ferror(ofd)) {
+	rpmlog(RPMLOG_ERR, _("rpmMkTemp failed\n"));
+	goto exit;
+    }
+
+    /* Write the lead/signature of the output rpm */
+    rc = rpmLeadWrite(ofd, lead);
+    if (rc != RPMRC_OK) {
+	rpmlog(RPMLOG_ERR, _("%s: writeLead failed: %s\n"), trpm,
+	    Fstrerror(ofd));
+	goto exit;
+    }
+
+    if (rpmWriteSignature(ofd, sigh)) {
+	rpmlog(RPMLOG_ERR, _("%s: rpmWriteSignature failed: %s\n"), trpm,
+	    Fstrerror(ofd));
+	goto exit;
+    }
+
+    /* Append the header and archive from the temp file */
+    if (copyFile(&fd, sigtarget, &ofd, trpm) == 0) {
+	struct stat st;
+
+	/* Move final target into place, restore file permissions. */
+	if (stat(rpm, &st) == 0 && unlink(rpm) == 0 &&
+		    rename(trpm, rpm) == 0 && chmod(rpm, st.st_mode) == 0) {
+	    res = 0;
+	} else {
+	    rpmlog(RPMLOG_ERR, _("replacing %s failed: %s\n"),
+		   rpm, strerror(errno));
+	}
+    }
 
 exit:
     if (fd)	(void) closeFile(&fd);
     if (ofd)	(void) closeFile(&ofd);
 
-    sigh = rpmFreeSignature(sigh);
+    rpmFreeSignature(sigh);
+    rpmLeadFree(lead);
 
     /* Clean up intermediate target */
     if (sigtarget) {
-	xx = unlink(sigtarget);
-	sigtarget = _free(sigtarget);
+	unlink(sigtarget);
+	free(sigtarget);
     }
     if (trpm) {
 	(void) unlink(trpm);
